/*
 * Administration routines for exos ethernet controller running a port
 * of the berkeley tcp/ip code.
 *
 * Written by: Kipp Hickman
 */

#ifdef SYSTEM5
#include "sys/param.h"
#include "sys/config.h"
#include "sys/types.h"
#include "sys/systm.h"
#include "sys/dir.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/file.h"
#include "sys/tty.h"
#include "sys/termio.h"
#include "sys/conf.h"
#include "sys/buf.h"
#endif

#ifdef VERSION7
#include <sys/param.h>
#include <sys/dir.h>
#include <sys/user.h>
#include <sys/buf.h>
#include <sys/proc.h>
#include <sys/file.h>
#endif

#include <sys/extypes.h>
#include <sys/socket.h>
#include <sys/swap.h>
#include <sys/exreg.h>
#include <sys/init.h>
#include <sys/exos.h>
#include <ex_ioctl.h>
#include "config.h"

struct	rmsg_area *rmsgarea;		/* read queues */
struct	wmsg_area *wmsgarea;		/* write queues */
struct	exctrl ex_db;			/* software device structure */

/* exos error logger variables */
static	short ad_logon;			/* error logger on? */
static	struct buf *ad_logbufp;		/* pointer to log buffer */
static	char *ad_wp;			/* circular queue pointer for writing */
static	char *ad_rp;			/* circular queue pointer for reading */
static	char *ad_endp;			/* pointer to end of data buf */
static	short ad_logcount;			/* count of bytes in log buf */


#undef	DEBUG				/* random debugging (low noise) */
#undef	DEBUG_NOISE			/* more noise */
#undef	DEBUG_QUEUES			/* ...and even more noise */

/*ARGSUSED*/
exopen(dev, flag)
	dev_t dev;
	int flag;
{
	register int d = minor(dev);

	switch (d) {
	  case ADMIN_LOGGER:
		if (ad_logon) {
			u.u_error = EBUSY;
			return;
		}
		ad_logbufp = geteblk();
		ad_logcount = 0;
		ad_wp = ad_rp = ad_logbufp->b_un.b_addr;
		ad_endp = ad_rp + BSIZE;
		ad_logon = 1;
	  case ADMIN_FIND:
	  default:
		break;
	}
}

/*ARGSUSED*/
exclose(dev, flag)
	dev_t dev;
	int flag;
{
	register short d = minor(dev);
	register short i;

	switch (d) {
	  case ADMIN_LOGGER:
		ad_logon = 0;
		brelse(ad_logbufp);
		break;
	  default:
		for (i = 0; i < MAXDEVS; i++) {
			if (so_state[i] == SOS_RESERVED)
				so_state[i] = SOS_FREE;
		}
		break;
	}
}

/*
 * exread:
 *	- read something from the admin device
 */
/*ARGSUSED*/
exread(dev)
	dev_t dev;
{
	register short d = minor(dev);
	register int count;
	register int s;

	switch (d) {
	  case ADMIN_LOGGER:
		while (ad_logcount == 0) {
			ad_logbufp->b_flags |= B_WANTED;
			sleep((caddr_t)ad_logbufp, PZERO+1);
		}

		s = spl7();
		if (ad_wp > ad_rp)
			iomove(ad_rp, count = (ad_wp - ad_rp), B_READ);
		else
			iomove(ad_rp, count = ad_endp - ad_rp, B_READ);
		ad_logcount -= count;
		if ((ad_rp += count) >= ad_endp)
			ad_rp = ad_logbufp->b_un.b_addr;
		splx(s);
		break;
	  default:
		u.u_error = EIO;
		break;
	}
}

/*ARGSUSED*/
exwrite(dev)
	dev_t dev;
{
	u.u_error = EIO;
}

/*
 * exintr:
 *	- process exos interrupts
 */
exintr(level)
	int level;
{
	register struct msg *read_mp;
	register struct Sock_printf *sp;
	register struct kludge *kp;

	outb(PORTA, 1);			/* clear interrupt */
#ifdef	DEBUG_NOISE
printf("[ex: ");
#endif
	read_mp = rmsgarea->ma_lastr;
	while ((read_mp->msg_hd.mh_status & (MQ_DONE|MQ_EXOS)) == 0) {
#ifdef	DEBUG_QUEUES
printf("%x %x %x ", read_mp->msg_hd.mh_status,
	     read_mp->msg_msg.nm_reply&0xff, read_mp->msg_msg.nm_request&0xff);
#endif
		kp = (struct kludge *)read_mp->msg_msg.nm_userid;
		switch (read_mp->msg_msg.nm_request&0xff) {
		  default:
			printf("funny message type %d\n",
				      read_mp->msg_msg.nm_request&0xff);
			if ((kp >= &kludge[0]) ||
			    (kp < &kludge[NET_KLUDGES]))
				wakeup((caddr_t)kp);
			break;
		  case SOSOCKET:
			kp->kl_soid = read_mp->msg_msg.nm_soid;
			goto finish;
		  case SOACCEPT:
		  case SOCONNECT:
#ifdef PEGASUS
		  case SOSOCKADDR:
#else
		  case SOSOCKETADDR:
#endif
			if (read_mp->msg_msg.nm_u.nm_cmd.nm_isaddr)
/* non portable
				kp->kl_sa =
				    read_mp->msg_msg.nm_u.nm_cmd.nm_saddr;
*/
				bcopy(&read_mp->msg_msg.nm_u.nm_cmd.nm_saddr,
					&kp->kl_sa,sizeof(struct sockaddr));
			goto finish;

		  case SORECEIVE:
			if (read_mp->msg_msg.nm_u.nm_packet.nm_isaddr)
/* non portable
				kp->kl_sa =
				    read_mp->msg_msg.nm_u.nm_packet.nm_saddr;
*/
				bcopy(&read_mp->msg_msg.nm_u.nm_packet.nm_saddr,
					&kp->kl_sa,sizeof(struct sockaddr));
			kp->kl_count =
				read_mp->msg_msg.nm_u.nm_packet.nm_count;
			/* FALL THROUGH to default */
		  case SOSEND:
		  case SOVERIFY:
		  case SOCLOSE:
		  case NET_DLOAD:
		  case NET_START:
finish:
			kp->kl_reply = read_mp->msg_msg.nm_reply&0xff;
			if (kp->kl_state & KL_WAITING) {
				kp->kl_state &= ~KL_WAITING;
				wakeup((caddr_t)kp);
			}
#ifdef	DEBUG_QUEUES
printf("wakeup(%x)\n", kp);
#endif	DEBUG_QUEUES
			break;
		  case NET_PANIC:
			sp = &read_mp->msg_msg.nm_u.nm_printf;
			ex_db.ex_netdown = 1;
			bcopy("exos panic\n", sp->nm_prdata, 12);
			/* FALL THROUGH */
		  case NET_PRINTF:
			sp = &read_mp->msg_msg.nm_u.nm_printf;
#ifdef DEBUG
printf("exintr: Got interrupt for net_printf: %s\n",sp->nm_prdata);
#endif
			exlog(sp->nm_prdata);
			break;
		  case IM_ALIVE:
			ex_db.ex_netdown = 0;
			printf("exos is alive and kicking\n");
			break;
		}
		read_mp->msg_hd.mh_status = MQ_EXOS|MQ_DONE;
		read_mp = read_mp->msg_link;
	}
	rmsgarea->ma_lastr = read_mp;
	if (ex_db.ex_state & ST_WAITING)
		wakeup((caddr_t)&ex_db);
#ifdef	DEBUG_NOISE
printf(" OK]\n");
#endif
	outb(PORTB, 0);
}

/*
 * exioctl:
 *	- process exos ioctl's
 */
/*ARGSUSED*/
exioctl(dev, cmd, addr, flag)
	dev_t dev;
	int cmd, flag;
	caddr_t addr;
{
	register short i;
	int socketdev;

#ifdef	DEBUG_NOISE
	printf("exioctl: cmd=%d\n", cmd);
#endif
	switch (cmd) {
	  case EXIOCFIND:
		for (i = 0; i < MAXDEVS; i++) {
			if (so_state[i] == SOS_FREE) {
				so_state[i] = SOS_RESERVED;
				socketdev = i;
#ifdef DEBUG
	printf("exioctl: EXIOCFIND returns device %d\n",socketdev);
#endif
				if (flag & FKERNEL)
					bcopy (&socketdev, addr,
						sizeof(socketdev));
				else
					if (copyout((caddr_t)&socketdev, addr,
					    	sizeof(socketdev)))
						u.u_error = EFAULT;
				return;
			}
		}
	  default:
		u.u_error = ENXIO;
	}
}

/*
 * exlog:
 *	- used by exos printf to output a string to the log
 *	- drop characters that won't fit in the buffer
 *	- SHOULD BYTE SWAP DATA HERE!
 */
exlog(msg)
	register char *msg;
{
	register int s;

	if (ad_logon) {
		s = spl7();
		while (ad_logcount < BSIZE) {
			if (*msg == 0)
				break;
			*ad_wp++ = *msg++;
			ad_logcount++;
			if (ad_wp == ad_endp)
				ad_wp = ad_logbufp->b_un.b_addr;
		}
		wakeup((caddr_t)ad_logbufp);
		splx(s);
	} else
		printf("%s", msg);
}
